import logging
import shutil
import tempfile
from .base import TargetBase
from git import Repo

LOG = logging.getLogger(__name__)
DEFAULT_MESSAGE = "updating repo programatically (via surgen)"

class GitTarget(TargetBase):
    def __init__(self, target, message=None, commit=True):
        super(GitTarget, self).__init__(target)
        self._message = message
        self._workspace = None
        self._commit = commit

    @property
    def workspace(self):
        return self._workspace

    def prepare(self):
        self._workspace = tempfile.mkdtemp()
        self.log("cloning {0}...".format(self.workspace))
        self._repo = Repo.clone_from(self._target, self.workspace)

    def commit(self, summary):
        """
        summary: a summary of the action performed on the target.
        """
        if not self._commit:
            self.log("committing disabled for {0}".format(self))
            return

        message = "{0}\n\n{1}".format(self._message or DEFAULT_MESSAGE, str(summary))

        try:
            self._repo.git.add(".")
            if self._repo.is_dirty():
                self.log("committing changes to {0}".format(self._target))
                self._repo.index.commit(message)
                self._repo.git.push()
            else:
                self.log("no changes found. skipping commit...")
        except Exception as e:
            LOG.exception("")
            print(e)
            self.log("unable to push to {0}".format(self._target))

    def cleanup(self):
        shutil.rmtree(self._workspace)



import os
from collections import OrderedDict
from clint.textui import colored, puts, indent
from .procedure import from_file
from .exceptions import ProcedureNotFound
from .results import ResultStatus, Result, Results

import logging

LOG = logging.getLogger(__name__)


class Surgen(object):
    """ the main context to perform operations. """

    def __init__(self, procedures_by_name):
        self._procedures_by_name = procedures_by_name
        # lists data about performed procedures
        self._performed_procedure_data = []

    def operate(self, target, ignore_errors, dry_run):
        """
        perform the operation on the target directory.
        return an exit code.
        """
        puts("Perfoming procedures on {0}".format(target))
        target.prepare()
        try:
            results = Results()
            with indent(2):
                for name, procedure_cls in self._procedures_by_name.items():
                    puts("{0}:".format(name))
                    with indent(2):
                        status = self._run_procedure(
                            name, procedure_cls, target.workspace, dry_run
                        )
                        results.add(Result(procedure=name, status=status))
                        if not ignore_errors and status == ResultStatus["FAIL"]:
                            with indent(-4):
                                puts(colored.red("Surgen ending early"))
                                self._print_results(results)
                            return 1
            if not dry_run:
                target.commit(results)
            self._print_results(results)
            self._performed_procedure_data.append(
                (target, results)
            )
            return 1 if ResultStatus["FAIL"] in results else 0
        finally:
            target.cleanup()

    def _print_results(self, results):
        puts(colored.green("Complete! {0} procedure(s) performed.".format(len(self._procedures_by_name))))
        puts(colored.green("{PASS} success, {FAIL} failed, {SKIP} skipped".format(**results.count_by_status)))

    def _run_procedure(self, name, procedure_cls, target_dir, dry_run):
        puts(colored.yellow("executing...".format(name)))
        procedure = procedure_cls(name, target_dir)
        should_not_run_reason = procedure.should_not_run()
        if should_not_run_reason:
            puts(colored.yellow("skipping, should_not_run returned: {0}".format(should_not_run_reason)))
            return ResultStatus["SKIP"]
        if not dry_run:
            try:
                with indent(2):
                    result = procedure.operate()
            except Exception as e:
                LOG.debug("", exc_info=True)
                puts(colored.red("procedure raised an exception! {0}".format(e)))
                return ResultStatus["FAIL"]
            puts(colored.green("complete!"))
        return ResultStatus["PASS"]

    def print_total_summary(self):
        """ print the total summary. """
        for target, results in self._performed_procedure_data:
            puts(colored.green("{0}: {PASS} | {FAIL} | {SKIP}".format(target, **results.count_by_status)))


def surgen_from_directory(procedure_dir):
    """ load a surgen object from a directory """
    return Surgen(_procedures_from_dir(procedure_dir))


def _procedures_from_dir(procedure_dir):
    procedures_by_name = OrderedDict()
    for d in os.listdir(procedure_dir):
        if os.path.isdir(d):
            puts(colored.yellow("skipping {0}, it is a directory."))
            continue
        if not d.endswith(".py"):
            puts(colored.yellow("skipping {0}, it does not end with extension .py".format(d)))
            continue
        name = d[:-len(".py")]
        full_path = os.path.join(procedure_dir, d)
        try:
            procedures_by_name[name] = from_file(full_path)
        except ProcedureNotFound as pnf:
            puts(colored.yellow("skipping {0}: {1}".format(d, pnf)))
            continue
    return procedures_by_name



import pytest
from surgen.procedure import from_string
from surgen.exceptions import ProcedureNotFound


def test_scope_works(tmpdir):
    """
    scripts which define variables in the global scope should
    have access to them.
    """

    SCRIPT = """
import os
from surgen import Procedure

class Scope(Procedure):

    def operate(self):
        print(os.path)
        return __file__

    """.strip()
    cls = from_string(SCRIPT, "foo")
    p = cls("foo", tmpdir.strpath)
    assert p.operate() == "foo"


def test_procedure_not_found_works(tmpdir):
    """
    no procedure class should raise an exception.
    """
    SCRIPT = """
import os
from surgen import Procedure
    """.strip()
    with pytest.raises(ProcedureNotFound):
        cls = from_string(SCRIPT, "foo")



from .exceptions import ProcedureNotFound
from clint.textui import puts


class Procedure(object):
    """
    the procedure represents the base class that a procedure file
    should extend from.

    it provides several utilities to help interact with a repository.
    """

    def __init__(self, file_name, target_dir):
        self._target_dir = target_dir
        self._file_name = file_name

    @property
    def root(self):
        return self._target_dir

    def should_not_run(self):
        """
        This should return None if the test should execute.
        otherwise, it should provide a string indicating the
        reason why it was skipped.
        """
        return None

    def log(self, message):
        """ log a message. """
        puts(message)

    def operate(self):
        """
        This method should be overriden, with the actual script.

        In the case of an error, an exception should be raised.
        """
        print("foo")

    def __str__(self):
        return self._file_name


def from_file(full_file_path):
    """ load a procedure from a full file path. """
    with open(full_file_path) as fh:
        contents = fh.read()
        return from_string(contents, full_file_path)


def from_string(string, file_name):
    l = {}
    l["__file__"] = file_name
    bytecode = compile(string, file_name, "exec")
    exec(bytecode, l, l)
    for name, value in l.items():
        if (
            isinstance(value, type) and
            issubclass(value, Procedure) and
            value is not Procedure
        ):
            return value
    raise ProcedureNotFound("unable to find class that inherits surgen.Procedure in file {0}".format(
        file_name
    ))
